from kombu.message import Message
from mock import Mock, patch, ANY, DEFAULT
import pytest

from nameko.constants import AMQP_URI_CONFIG_KEY
from nameko.containers import ServiceContainer, WorkerContext
from nameko.exceptions import RemoteError, ContainerBeingKilled
from nameko_nova_compat.extensions import (
    rpc, NovaRpc, NovaResponder, NovaRpcConsumer)
from nameko_nova_compat.proxy import RPCProxy


class NovaService(object):
    name = "novaservice"

    @rpc
    def echo(self, arg):
        return arg


@pytest.yield_fixture
def mock_publish():
    path = 'nameko_nova_compat.extensions.producers'
    with patch(path) as patched:
        publish = patched[ANY].acquire().__enter__().publish
        yield publish


def test_nova_rpc(container_factory, rabbit_config):

    container = container_factory(NovaService, rabbit_config)
    container.start()

    uri = rabbit_config['AMQP_URI']
    proxy = RPCProxy(uri)

    assert proxy.novaservice.echo(arg="hello") == "hello"

    with pytest.raises(RemoteError) as exc:
        proxy.novaservice.not_a_method()
    assert "MethodNotFound" in exc.value.message

    with pytest.raises(RemoteError) as exc:
        proxy.novaservice.echo()
    assert "IncorrectSignature" in exc.value.message

    container.stop()


def test_nova_rpc_provider(empty_config):

    rpc_consumer = Mock()
    message = Mock(headers={})

    message_body = {
        'method': 'method',
        'args': {"arg": "arg_value"},
        'msg_id': 'msg_id',
        '_context_user_id': 'user_id'
    }

    class Service(object):
        def method(self, arg):
            pass

    container = Mock(spec=ServiceContainer)
    container.shared_extensions = {}
    container.service_cls = Service
    container.worker_ctx_cls = WorkerContext
    container.service_name = "service"
    container.config = empty_config

    entrypoint = NovaRpc().bind(container, "method")
    entrypoint.setup()
    entrypoint.rpc_consumer = rpc_consumer

    container.spawn_worker.side_effect = ContainerBeingKilled()
    entrypoint.handle_message(message_body, message)
    assert rpc_consumer.requeue_message.called


def test_nova_responder(mock_publish):

    config = {AMQP_URI_CONFIG_KEY: ''}
    responder = NovaResponder(config, "msgid")

    # serialisable result
    result, exc_info = responder.send_response(True, None)
    assert result is True
    assert exc_info is None

    assert mock_publish.call_count == 2
    data_call, marker_call = mock_publish.call_args_list
    (data_msg,), _ = data_call
    (marker_msg,), _ = marker_call

    assert data_msg == {
        'failure': None,
        'result': True,
        'ending': False
    }
    assert marker_msg == {
        'failure': None,
        'result': None,
        'ending': True
    }


def test_nova_responder_unserializale_result(mock_publish):

    config = {AMQP_URI_CONFIG_KEY: ''}
    responder = NovaResponder(config, "msgid")

    # unserialisable result
    obj = object()
    result, exc_info = responder.send_response(obj, None)
    assert result is None
    assert exc_info == (TypeError, ANY, ANY)

    assert mock_publish.call_count == 2
    data_call, _ = mock_publish.call_args_list
    (data_msg,), _ = data_call

    assert data_msg == {
        'failure': ('TypeError', "{} is not JSON serializable".format(obj)),
        'result': None,
        'ending': False
    }


def test_nova_responder_cannot_str_exc(mock_publish):

    config = {AMQP_URI_CONFIG_KEY: ''}
    responder = NovaResponder(config, "msgid")

    class BadException(Exception):
        def __str__(self):
            raise Exception('boom')

    # un-str-able exception
    exc = BadException()
    result, exc_info = responder.send_response(
        True, (BadException, exc, "tb"))
    assert result is True
    assert exc_info == (BadException, exc, "tb")

    assert mock_publish.call_count == 2
    data_call, _ = mock_publish.call_args_list
    (data_msg,), _ = data_call

    assert data_msg == {
        'failure': ('BadException', "[__str__ failed]"),
        'result': True,
        'ending': False
    }


def test_nova_consumer_bad_provider():
    container = Mock(spec=ServiceContainer)
    container.shared_extensions = {}

    consumer = NovaRpcConsumer().bind(container)
    message = Message(
        channel=None,
        delivery_info={'routing_key': 'some route'},
        properties={},
    )
    with patch.multiple(
        consumer,
        get_provider_for_method=DEFAULT,
        handle_result=DEFAULT,
    ) as mocks:
        provider = mocks['get_provider_for_method']
        handle_result = mocks['handle_result']

        exception = LookupError('broken')
        provider.side_effect = exception
        consumer.handle_message({'args': ()}, message)
        assert handle_result.call_count == 1
        args, kwargs = handle_result.call_args
        exc_info = args[-1]
        exc_type, exc_value, traceback = exc_info
        assert exc_value is exception



from mock import ANY, Mock, patch
import pytest

from nameko_nova_compat import proxy
from nameko_nova_compat.testing import MockRPCProxy


def test_anon_context_constructor():
    context = proxy.get_anon_context()

    assert context.user_id is None


@pytest.mark.parametrize('constructor', [proxy.RPCProxy, MockRPCProxy])
@patch.object(proxy, 'nova', autospec=True)
def test_call(nova, constructor):
    connection = ANY
    context = Mock()
    rpcproxy = constructor(context_factory=lambda: context)

    with pytest.raises(ValueError):
        # no topic, no method
        rpcproxy.call(key='value')

    with pytest.raises(ValueError):
        # no method
        rpcproxy.service.call(key='value')

    rpcproxy.service.controller.call(key='value')

    nova.call.assert_called_once_with(
        connection, context, 'service',
        {'method': 'controller', 'args': {'key': 'value'}},
        options=rpcproxy.call_options(),
        timeout=None)


@pytest.mark.parametrize('constructor', [proxy.RPCProxy, MockRPCProxy])
@patch.object(proxy, 'nova', autospec=True)
def test_call_dynamic_route(nova, constructor):
    connection = ANY
    context = Mock()
    rpcproxy = constructor(context_factory=lambda: context)

    with pytest.raises(ValueError):
        # no topic, no method
        rpcproxy.call(key='value')

    with pytest.raises(ValueError):
        # no method
        rpcproxy.call(topic='service', key='value')

    with pytest.raises(ValueError):
        # no topic
        rpcproxy.call(method='controller', key='value')

    rpcproxy.call(topic='service', method='controller', key='value')

    nova.call.assert_called_once_with(
        connection, context, 'service',
        {'method': 'controller', 'args': {'key': 'value'}},
        options=rpcproxy.call_options(),
        timeout=None)


@pytest.mark.parametrize('constructor', [proxy.RPCProxy, MockRPCProxy])
@patch.object(proxy, 'nova', autospec=True)
def test_call_default(nova, constructor):
    connection = ANY
    context = Mock()
    rpcproxy = constructor(context_factory=lambda: context)

    rpcproxy.service.controller(key='value')

    nova.call.assert_called_once_with(
        connection, context, 'service',
        {'method': 'controller', 'args': {'key': 'value'}},
        options=rpcproxy.call_options(),
        timeout=None)


@pytest.mark.parametrize('constructor', [proxy.RPCProxy, MockRPCProxy])
@patch.object(proxy, 'nova', autospec=True)
def test_extra_route(nova, constructor):
    rpcproxy = constructor()

    with pytest.raises(AttributeError):
        rpcproxy.service.controller.extra(key='value')


@pytest.mark.parametrize('constructor', [proxy.RPCProxy, MockRPCProxy])
@patch.object(proxy, 'nova', autospec=True)
def test_route_abuse(nova, constructor):
    rpcproxy = constructor()
    # N.B. There are safeguards against misconfiguring the info attribute.
    #      If it's got to this point someone has been misusing the api.
    rpcproxy.info = ['service', 'controller', 'extra']

    with pytest.raises(ValueError):
        rpcproxy.call(key='value')


@pytest.mark.parametrize('constructor', [proxy.RPCProxy, MockRPCProxy])
@patch.object(proxy, 'nova', autospec=True)
def test_control_exchange_config(rpc, constructor):
    connection = ANY
    context = Mock()
    rpcproxy = constructor(
        control_exchange='rpc', context_factory=lambda: context)

    rpcproxy.service.controller(key='value')

    rpc.call.assert_called_once_with(
        connection, context, 'service',
        {'method': 'controller', 'args': {'key': 'value'}},
        options={'CONTROL_EXCHANGE': 'rpc'},
        timeout=None)


# MockRPCProxy Tests

def test_add_dummy():
    rpcproxy = MockRPCProxy()
    rpcproxy.fallback_to_call = False

    with pytest.raises(RuntimeError):
        rpcproxy.service.controller()

    response = Mock()
    rpcproxy.add_dummy('service', 'controller', response)

    assert rpcproxy._calls == [
        ('service', 'controller', {}),
    ]

    assert rpcproxy.service.controller() == response
    assert rpcproxy.service.controller(key='value') == response

    assert rpcproxy._calls == [
        ('service', 'controller', {}),
        ('service', 'controller', {}),
        ('service', 'controller', {'key': 'value'}),
    ]

    with pytest.raises(RuntimeError):
        rpcproxy.service.other_controller()

    assert rpcproxy._calls == [
        ('service', 'controller', {}),
        ('service', 'controller', {}),
        ('service', 'controller', {'key': 'value'}),
        ('service', 'other_controller', {}),
    ]

    rpcproxy.reset()
    rpcproxy.fallback_to_call = False

    assert rpcproxy._calls == []

    with pytest.raises(RuntimeError):
        rpcproxy.service.controller()


def test_add_call_matching():
    rpcproxy = MockRPCProxy()
    rpcproxy.fallback_to_call = False

    with pytest.raises(RuntimeError):
        rpcproxy.service.controller()

    response = Mock()
    rpcproxy.add_call_matching(
        'service', 'controller', {'key': 'value'}, response)

    assert rpcproxy._calls == [
        ('service', 'controller', {}),
    ]

    with pytest.raises(RuntimeError):
        rpcproxy.service.controller()

    assert rpcproxy.service.controller(key='value') == response

    assert rpcproxy._calls == [
        ('service', 'controller', {}),
        ('service', 'controller', {}),
        ('service', 'controller', {'key': 'value'}),
    ]

    with pytest.raises(RuntimeError):
        rpcproxy.service.other_controller()

    assert rpcproxy._calls == [
        ('service', 'controller', {}),
        ('service', 'controller', {}),
        ('service', 'controller', {'key': 'value'}),
        ('service', 'other_controller', {}),
    ]

    rpcproxy.reset()
    rpcproxy.fallback_to_call = False

    assert rpcproxy._calls == []

    with pytest.raises(RuntimeError):
        rpcproxy.service.controller(key='value')


@patch.object(proxy, 'nova', autospec=True)
def test_service_whitelist(nova):
    connection = ANY
    context = Mock()
    rpcproxy = MockRPCProxy(context_factory=lambda: context)
    rpcproxy.fallback_to_call = False

    with pytest.raises(RuntimeError):
        rpcproxy.service.controller()

    assert nova.call.call_count == 0

    rpcproxy.add_service_to_whitelist('service')

    rpcproxy.service.controller()

    nova.call.assert_called_once_with(
        connection, context, 'service',
        {'method': 'controller', 'args': {}},
        options=rpcproxy.call_options(),
        timeout=None)

    with pytest.raises(RuntimeError):
        rpcproxy.blacklisted_service.controller()

    rpcproxy.reset()
    rpcproxy.fallback_to_call = False

    with pytest.raises(RuntimeError):
        rpcproxy.service.controller()

    assert nova.call.call_count == 1


def test_fallback_to_call():
    rpcproxy = MockRPCProxy()
    assert rpcproxy.fallback_to_call is True  # default
    rpcproxy.fallback_to_call = False
    assert rpcproxy.fallback_to_call is False


@pytest.mark.parametrize('constructor', [proxy.RPCProxy, MockRPCProxy])
@patch.object(proxy, 'nova', autospec=True)
def test_timeout(nova, constructor):
    connection = ANY
    context = Mock()

    def context_factory():
        return context

    # test null timeout
    rpcproxy = constructor(context_factory=context_factory)
    rpcproxy.service.controller(key='value')

    nova.call.assert_called_once_with(
        connection, context, 'service',
        {'method': 'controller', 'args': {'key': 'value'}},
        options=rpcproxy.call_options(),
        timeout=None)
    nova.reset_mock()

    # test default timeout
    rpcproxy = constructor(timeout=10, context_factory=context_factory)
    rpcproxy.service.controller(key='value')

    nova.call.assert_called_once_with(
        connection, context, 'service',
        {'method': 'controller', 'args': {'key': 'value'}},
        options=rpcproxy.call_options(),
        timeout=10)
    nova.reset_mock()

    # test override timeout
    rpcproxy = constructor(timeout=10, context_factory=context_factory)
    rpcproxy.service.controller(key='value', timeout=99)

    nova.call.assert_called_once_with(
        connection, context, 'service',
        {'method': 'controller', 'args': {'key': 'value'}},
        options=rpcproxy.call_options(),
        timeout=99)



import types


class ChannelHandler(object):
    def __init__(self, connection, channel=None, create_channel=True):
        self.connection = connection
        if channel is None and create_channel:
            self.channel = connection.channel()
        else:
            self.channel = channel

    def close(self):
        if self.channel is not None:
            self.channel.close()

    def revive(self, channel):
        self.channel = channel

    def __call__(self, func, *args, **kwargs):
        return self.ensure(func)(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_typ, exc_val, exc_tb):
        self.close()

    def ensure(self, func):
        if isinstance(func, types.MethodType):
            obj = func.im_self
        elif isinstance(func, tuple):
            obj, func = func
        else:
            raise TypeError('could not obtain object from function!')

        if obj is self:
            revive = self.revive
        else:
            revive = None

        return self.connection.ensure(obj, func, on_revive=revive)



import iso8601

from nameko_nova_compat.common import UTCNOW, UIDGEN


class Context(dict):

    def __init__(self, user_id, timestamp=None, request_id=None, **kwargs):
        if isinstance(timestamp, basestring):
            timestamp = iso8601.parse_date(timestamp)
        self.timestamp = timestamp or UTCNOW()
        self.request_id = request_id or UIDGEN()
        self.update(kwargs)

        # we currently care little about what goes in here
        # from nova's context
        self.user_id = user_id

    def __getattr__(self, name):
        if name not in self:
            raise AttributeError("Context has no attribute '{}'".format(name))
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]

    def __repr__(self):
        return 'Context({})'.format(self.to_dict())

    def to_dict(self):
        # TODO: convert to UTC if not?
        res = self.copy()

        timestamp = self.timestamp.isoformat()
        timestamp = timestamp[:timestamp.index('+')]
        res.update({
            'timestamp': timestamp,
            'project_id': None,
        })
        return res

    def add_to_message(self, message):
        return add_context_to_payload(self, message)


def get_admin_context():
    return Context(user_id=None, is_admin=True)


def add_context_to_payload(context, payload):
    for key, value in context.to_dict().iteritems():
        payload['_context_{}'.format(key)] = value
    return payload



import mock
import pytest

from nameko_nova_compat.channelhandler import ChannelHandler


def test_ensure():
    conn = mock.Mock()

    handler = ChannelHandler(conn, create_channel=False)

    handler.ensure(handler.close)
    conn.ensure.assert_called_with(
        handler, handler.close, on_revive=handler.revive)

    obj = {}

    def fn():
        return None

    handler.ensure((obj, fn))
    conn.ensure.assert_called_with(obj, fn, on_revive=None)

    with pytest.raises(TypeError):
        handler.ensure(None)



import unittest

from nameko_nova_compat import context
from nameko_nova_compat.nova import parse_message


class TestContext(unittest.TestCase):

    def _makeOne(self, *args, **kwargs):
        return context.Context(*args, **kwargs)

    def test_constructor_defaults(self):
        user_id = 1
        ctx = self._makeOne(user_id)

        assert ctx.user_id == ctx['user_id'] == user_id

        assert ctx.timestamp == ctx['timestamp']
        assert ctx.timestamp is not None

        assert ctx.request_id == ctx['request_id']
        assert ctx.request_id is not None

    def test_custom_constructor_args(self):
        user_id = 1
        ctx = self._makeOne(user_id, custom='foobar')

        assert ctx.user_id == ctx['user_id'] == user_id

        assert ctx.custom == ctx['custom'] == 'foobar'

    def test_repr(self):
        user_id = 1
        ctx = self._makeOne(user_id)

        assert repr(ctx) == 'Context({})'.format(ctx.to_dict())

    def test_attribute_behaviour(self):
        user_id = 1
        ctx = self._makeOne(user_id)

        with self.assertRaises(AttributeError):
            ctx.custom

        ctx.custom = 'foobar'
        assert ctx.custom == 'foobar'

        del ctx.custom

        with self.assertRaises(AttributeError):
            ctx.custom

    def test_dict_behaviour(self):
        user_id = 1
        ctx = self._makeOne(user_id)

        assert 'custom' not in ctx
        assert ctx.get('custom') is None
        with self.assertRaises(KeyError):
            ctx['custom']

        ctx['custom'] = 'foobar'

        assert 'custom' in ctx
        assert ctx.get('custom') == 'foobar'
        assert ctx['custom'] == 'foobar'

        del ctx['custom']

        assert 'custom' not in ctx
        assert ctx.get('custom') is None
        with self.assertRaises(KeyError):
            ctx['custom']

    def test_to_dict(self):
        """ context should serialise the timestamp for transport
        """
        from datetime import datetime

        user_id = 1
        ctx = self._makeOne(user_id)

        assert isinstance(ctx['timestamp'], datetime)

        serialised = ctx.to_dict()

        assert isinstance(serialised['timestamp'], str)

    def test_transport(self):
        user_id = 1
        ctx = self._makeOne(user_id)

        message_base = {
            'method': 'rpc_call',
            'args': {'arg1': 'val1'},
            '_msg_id': ctx.request_id,
        }
        message = message_base.copy()

        ctx.add_to_message(message)

        msg_id, new_ctx, method, args = parse_message(message)

        assert msg_id == message_base['_msg_id']
        assert id(ctx) != id(new_ctx)
        assert ctx.to_dict() == new_ctx.to_dict()
        assert method == message_base['method']
        assert args == message_base['args']



from __future__ import absolute_import
from logging import getLogger

from kombu import Producer, Exchange, Queue
from nameko.constants import DEFAULT_RETRY_POLICY
from nameko.exceptions import UnknownService
from nameko.kombu_helpers import queue_iterator

from nameko_nova_compat import responses
from nameko_nova_compat.context import Context
from nameko_nova_compat.channelhandler import ChannelHandler
from nameko_nova_compat.common import UIDGEN
from nameko_nova_compat.decorators import ensure


_log = getLogger(__name__)


CONTROL_EXCHANGE = 'rpc'
DEFAULT_RPC_TIMEOUT = 10
DURABLE_QUEUES = False


def get_reply_exchange(msgid, channel=None):
    return Exchange(
        name=msgid,
        channel=channel,
        type='direct',
        durable=False,
        auto_delete=True)


def get_reply_queue(msgid, channel=None):
    exchange = get_reply_exchange(msgid, channel=channel)
    return Queue(
        name=msgid,
        channel=channel,
        exchange=exchange,
        routing_key=msgid,
        durable=False,
        auto_delete=True,
        exclusive=True)


def get_topic_exchange(exchange_name, channel=None):
    return Exchange(
        name=exchange_name,
        channel=channel,
        type='topic',
        durable=DURABLE_QUEUES,
        auto_delete=False)


def get_topic_queue(exchange_name, topic, channel=None):
    exchange = get_topic_exchange(exchange_name, channel=channel)
    return Queue(
        name=topic,
        channel=channel,
        exchange=exchange,
        routing_key=topic,
        durable=DURABLE_QUEUES,
        auto_delete=False,
        exclusive=False)


def _get_exchange(options):
    if options is not None:
        return options.get('CONTROL_EXCHANGE', CONTROL_EXCHANGE)
    return CONTROL_EXCHANGE


def _create_rpcpayload(context, method, args):
    message = {'method': method, 'args': args, }
    message = context.add_to_message(message)
    msg_id = UIDGEN()
    message['_msg_id'] = msg_id
    return msg_id, message


def _send_topic(connection, exchange, topic, data):
    exchange = get_topic_exchange(exchange)
    with ChannelHandler(connection) as ch:
        producer = Producer(
            ch.channel,
            exchange=exchange,
            routing_key=topic,
        )
        ch.ensure(producer.publish)(data, declare=[exchange], mandatory=True,
                                    retry_policy=DEFAULT_RETRY_POLICY,
                                    retry=True)

        # see comment in rpc.MethodProxy.__call__, after publish()
        if not producer.channel.returned_messages.empty():
            raise UnknownService(topic)


def parse_message(message_body):
    method = message_body.get('method')
    args = message_body.get('args')
    msg_id = message_body.get('_msg_id', None)
    context_dict = dict(
        (k[9:], message_body.get(k))
        for k in message_body.keys() if k.startswith('_context_')
    )
    context = Context(**context_dict)
    return msg_id, context, method, args


@ensure
def send_rpc(connection, context, exchange, topic, method, args, timeout=None):

    _log.info('rpc: %s %s.%s', exchange, topic, method)

    msgid, payload = _create_rpcpayload(context, method, args)

    with connection.channel() as channel:
        queue = get_reply_queue(msgid, channel=channel)
        queue.declare()
        _send_topic(connection, exchange, topic, payload)
        iter_ = queue_iterator(queue, timeout=timeout)
        iter_ = (msg for (_, msg) in iter_)
        iter_ = responses.iter_rpcresponses(iter_)
        ret = responses.last(iter_)
        if ret is not None:
            return ret.payload['result']


def call(connection, context, topic, msg,
         timeout=DEFAULT_RPC_TIMEOUT, options=None):

    exchange = _get_exchange(options)

    return send_rpc(
        connection,
        context=context,
        exchange=exchange,
        topic=topic,
        method=msg['method'],
        args=msg['args'],
        timeout=timeout)




import functools

from nameko_nova_compat.channelhandler import ChannelHandler


def ensure(func):
    """wraps an entire function with ensure to use as a decorator. """
    @functools.wraps(func)
    def wrapped(connection, *args, **kwargs):
        with ChannelHandler(connection, create_channel=False) as ch:
            return ch((ch, func), connection, *args, **kwargs)
    return wrapped



from nameko_nova_compat import decorators


class _Tester(object):
    __name__ = 'Tester'  # for functools.wraps()

    def __init__(self, failtimes=1, failexc=IOError):
        self.failtimes = failtimes
        self.tries = 0
        self.failexc = failexc

    def __call__(self, channel, *args, **kwargs):
        self.tries += 1
        if self.tries <= self.failtimes:
            raise self.failexc()


def test_ensure(connection):
    tester = _Tester()
    connection.transport.connection_errors = (tester.failexc,)

    decorators.ensure(tester)(connection)
    assert tester.tries == 2



import kombu

from nameko_nova_compat.context import Context
from nameko_nova_compat import nova


def get_anon_context():
    return Context(user_id=None, project_id=1)


class RPCProxy(object):
    """ Supports calling other services using the format::

            >>> proxy = RPCProxy()
            >>> proxy.service_name.controller_name(arg1='foo', arg2='bar')
    """

    def __init__(self, uri='memory://', timeout=None, info=None,
                 context_factory=None, control_exchange=None):
        self.uri = uri
        self.timeout = timeout
        self.info = info or []
        self.context_factory = context_factory or get_anon_context
        self.control_exchange = control_exchange

    def create_connection(self):
        return kombu.BrokerConnection(
            self.uri,
            transport_options={'confirm_publish': True},
        )

    def call_options(self):
        options = {}
        if self.control_exchange:
            options['CONTROL_EXCHANGE'] = self.control_exchange
        return options

    def __getattr__(self, key):
        if len(self.info) >= 2:
            raise AttributeError(key)
        info = self.info[:] + [key]
        return self.__class__(
            uri=self.uri,
            timeout=self.timeout,
            info=info,
            context_factory=self.context_factory,
            control_exchange=self.control_exchange,
        )

    def _get_route(self, kwargs):
        info = self.info

        if len(info) == 2:
            topic, method = info

        elif len(info) < 2:
            if len(info) == 1:
                topic = info[0]
            else:
                topic = kwargs.pop('topic', None)
            method = kwargs.pop('method', None)

            if None in (topic, method):
                raise ValueError('topic and method need providing')

        elif len(info) > 2:
            raise ValueError(
                'Only topic and method should be specified. '
                'Got topic={}, method={}, extra={}'.format(
                    info[0], info[1], ', '.join(info[2:]))
            )

        return (topic, method)

    def call(self, context=None, **kwargs):
        topic, method = self._get_route(kwargs)
        timeout = kwargs.pop('timeout', self.timeout)

        if context is None:
            context = self.context_factory()

        with self.create_connection() as connection:
            return nova.call(
                connection, context, topic,
                {'method': method, 'args': kwargs, },
                timeout=timeout,
                options=self.call_options(),
            )

    def __call__(self, context=None, **kwargs):
        return self.call(context, **kwargs)



from nameko.exceptions import RemoteError


def ifirst(iter_):
    # TODO: this is rather confusing as it would return the same
    #       for an empty iterable and one with the first item being None.
    #       How would we know if it actually was the first item?
    for i in iter_:
        return i


def last(iter_):
    i = None
    prev = None
    for i in iter_:
        if prev is not None:
            prev.ack()
        prev = i
    i.ack()
    return i


def iter_rpcresponses(iter_):
    for msg in iter_:
        data = msg.payload
        if data['failure']:
            msg.ack()
            raise RemoteError(*data['failure'])
        elif data.get('ending', False):
            msg.ack()
            return
        else:
            yield msg
    # TODO: This might work even if we never have a data['ending'] == True
    #       and the sender closes it's connection prematurely,
    #       using memory transport. It should not! Needs further investigation.



from nameko_nova_compat.proxy import RPCProxy

_state = {}


def reset_state():
    _state['calls'] = []
    _state['routes'] = {}
    _state['services_whitelist'] = []
    _state['fallback_to_call'] = True

reset_state()


class MockRPCProxy(RPCProxy):
    def __init__(self, *args, **kwargs):
        super(MockRPCProxy, self).__init__(*args, **kwargs)

    @staticmethod
    def reset():
        reset_state()

    @staticmethod
    def add_routing(topic, method, func, priority=0):
        routes = _state['routes'].setdefault(topic, {}).setdefault(method, [])
        routes.append((priority, func))
        routes.sort(reverse=True)

    @staticmethod
    def add_dummy(topic, method, retvalue=None, priority=0):
        MockRPCProxy.add_routing(
            topic, method, lambda *args, **kwargs: retvalue,
            priority=priority)

    @staticmethod
    def add_call_matching(topic, method, args, returnvalue):
        def routefunc(context, **kwargs):
            assert kwargs == args
            return returnvalue
        MockRPCProxy.add_routing(topic, method, routefunc)

    @staticmethod
    def add_service_to_whitelist(service):
        _state['fallback_to_call'] = True
        _state['services_whitelist'].append(service)

    # keep self.fallback_to_call for backwards compatibility
    @property
    def fallback_to_call(self):
        return _state['fallback_to_call']

    @fallback_to_call.setter
    def fallback_to_call(self, value):
        _state['fallback_to_call'] = value

    # keep self._calls for backwards compatibility
    @property
    def _calls(self):
        return _state['calls']

    def __call__(self, context=None, **kwargs):
        topic, method = self._get_route(kwargs.copy())

        if context is None:
            context = self.context_factory()

        _state['calls'].append((topic, method, kwargs))

        # check the routes registered on the mock object first
        routes = _state['routes'].get(topic, {}).get(method)
        if routes:
            for _, r in routes:
                try:
                    return r(context, **kwargs)
                except (ValueError, AssertionError):
                    pass

        # no route was registered so fallback to a real call if it's allowed
        if _state['fallback_to_call']:
            whitelist = _state['services_whitelist']
            if whitelist and topic not in whitelist:
                raise RuntimeError(
                    'Service not in whitelist when trying '
                    'to call {}.{}(args={})'.format(topic, method, kwargs)
                )
            return self.call(context, **kwargs)

        # unable to process the requested route
        raise RuntimeError(
            'No Matching RPC Routes for {}.{}(args={})'.format(
                topic, method, kwargs)
        )



import eventlet
from eventlet.event import Event
from kombu import Producer
import mock
import pytest

from nameko.exceptions import RemoteError, UnknownService
from nameko.kombu_helpers import queue_iterator
from nameko_nova_compat import context
from nameko_nova_compat import nova
from nameko_nova_compat.responses import ifirst
from nameko.testing.utils import assert_stops_raising


def test_delegation_to_send_rpc():

    conn = 'connection'
    ctx = 'context'
    topic = 'topic'
    method = 'foobar'
    args = 'args'
    msg = dict(method=method, args=args)
    timeout = 123
    exchange = 'spam_exchange'
    options = {'CONTROL_EXCHANGE': exchange}

    with mock.patch(
        'nameko_nova_compat.nova.send_rpc', autospec=True
    ) as send_rpc:
        nova.call(
            connection=conn, context=ctx, topic=topic,
            msg=msg, timeout=timeout, options=options)

        send_rpc.assert_called_with(
            conn, context=ctx, exchange=exchange,
            topic=topic, method=method, args=args,
            timeout=timeout)


def test_delegation_to_send_rpc_default_exchange():

    conn = 'connection'
    ctx = 'context'
    topic = 'topic'
    method = 'foobar'
    args = 'args'
    msg = dict(method=method, args=args)
    timeout = 123
    exchange = 'rpc'

    with mock.patch(
        'nameko_nova_compat.nova.send_rpc', autospec=True
    ) as send_rpc:
        nova.call(
            connection=conn, context=ctx, topic=topic,
            msg=msg, timeout=timeout)

        send_rpc.assert_called_with(
            conn, context=ctx, exchange=exchange,
            topic=topic, method=method, args=args,
            timeout=timeout)


def test_send_rpc(get_connection):

    queue_declared = Event()

    def response_greenthread():
        with get_connection() as conn:
            with conn.channel() as chan:
                queue = nova.get_topic_queue(
                    'test_rpc', 'test', channel=chan)
                queue.declare()
                queue_declared.send(True)
                body, msg = ifirst(
                    queue_iterator(queue, no_ack=True, timeout=2))
                msgid, _, _, args = nova.parse_message(body)

                exchange = nova.get_reply_exchange(msgid)
                producer = Producer(chan, exchange=exchange, routing_key=msgid)

                msg = {'result': args, 'failure': None, 'ending': False}
                producer.publish(msg)
                msg = {'result': None, 'failure': None, 'ending': True}
                producer.publish(msg)

    g = eventlet.spawn_n(response_greenthread)
    eventlet.sleep(0)

    with get_connection() as conn:
        ctx = context.get_admin_context()

        queue_declared.wait()
        resp = nova.send_rpc(
            conn,
            context=ctx,
            exchange='test_rpc',
            topic='test',
            method='test_method',
            args={'foo': 'bar', },
            timeout=3)

        assert resp == {'foo': 'bar', }

    def check_greenthread_dead():
        assert not g
    assert_stops_raising(check_greenthread_dead)


def test_send_rpc_unknown_service(get_connection):
    with get_connection() as conn:
        ctx = context.get_admin_context()

        with pytest.raises(UnknownService):
            nova.send_rpc(
                conn,
                context=ctx,
                exchange='test_rpc',
                topic='test',
                method='test_method',
                args={'foo': 'bar', },
                timeout=3)


def test_send_rpc_errors(get_connection):

    queue_declared = Event()

    def response_greenthread():
        with get_connection() as conn:
            with conn.channel() as chan:
                queue = nova.get_topic_queue(
                    'test_rpc', 'test', channel=chan)
                queue.declare()
                queue_declared.send(True)
                body, msg = ifirst(
                    queue_iterator(queue, no_ack=True, timeout=2))
                msgid, _, _, _ = nova.parse_message(body)

                exchange = nova.get_reply_exchange(msgid)
                producer = Producer(chan, exchange=exchange, routing_key=msgid)

                exc = Exception('error')
                failure = (type(exc).__name__, str(exc))

                msg = {'result': None, 'failure': failure, 'ending': False}
                producer.publish(msg)
                msg = {'result': None, 'failure': None, 'ending': True}
                producer.publish(msg)

    g = eventlet.spawn_n(response_greenthread)
    eventlet.sleep(0)

    with get_connection() as conn:
        ctx = context.get_admin_context()

        with pytest.raises(RemoteError):

            queue_declared.wait()
            nova.send_rpc(
                conn,
                context=ctx,
                exchange='test_rpc',
                topic='test',
                method='test_method',
                args={'foo': 'bar', },
                timeout=3)

    def check_greenthread_dead():
        assert not g
    assert_stops_raising(check_greenthread_dead)


def test_send_rpc_multi_message_reply_ignores_all_but_last(get_connection):

    queue_declared = Event()

    def response_greenthread():
        with get_connection() as conn:
            with conn.channel() as chan:
                queue = nova.get_topic_queue(
                    'test_rpc', 'test', channel=chan)
                queue.declare()
                queue_declared.send(True)

                body, msg = ifirst(
                    queue_iterator(queue, no_ack=True, timeout=2))
                msgid, _, _, args = nova.parse_message(body)

                exchange = nova.get_reply_exchange(msgid)
                producer = Producer(chan, exchange=exchange, routing_key=msgid)

                for _ in range(3):
                    msg = dict(
                        result='should ignore this message',
                        failure=None, ending=False)
                    producer.publish(msg)
                    eventlet.sleep(0.1)

                msg = dict(result=args, failure=None, ending=False)
                producer.publish(msg)
                msg = dict(result=None, failure=None, ending=True)
                producer.publish(msg)

    g = eventlet.spawn_n(response_greenthread)
    eventlet.sleep()

    with get_connection() as conn:
        ctx = context.get_admin_context()

        queue_declared.wait()
        resp = nova.send_rpc(
            conn,
            context=ctx,
            exchange='test_rpc',
            topic='test',
            method='test_method',
            args={'spam': 'shrub', },
            timeout=3)

        assert resp == {'spam': 'shrub', }
    eventlet.sleep()

    def check_greenthread_dead():
        assert not g
    assert_stops_raising(check_greenthread_dead)



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(
    name='nameko-nova-compat',
    version='0.0.1',
    description='Nova-RPC compatibility for nameko services',
    author='onefinestay',
    author_email='engineering@onefinestay.com',
    url='http://github.com/onefinestay/nameko-nova-compat',
    packages=find_packages(exclude=['test']),
    install_requires=[
        "nameko>=2.0.0",
        "iso8601"
    ],
    extras_require={
        'dev': [
            "coverage==4.0a1",
            "flake8==2.1.0",
            "mccabe==0.3",
            "pep8==1.6.1",
            "pyflakes==0.8.1",
            "pylint==1.0.0",
            "pytest==2.4.2",
            "pytest-timeout==0.4",
        ]
    },
    dependency_links=[],
    zip_safe=True,
    license='Apache License, Version 2.0',
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.7",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
    ]
)



from functools import partial
import json
import sys

from kombu import Connection
from kombu.pools import producers
from nameko.constants import DEFAULT_RETRY_POLICY, AMQP_URI_CONFIG_KEY
from nameko.exceptions import ContainerBeingKilled
from nameko.rpc import RpcConsumer, Rpc, Responder

from nameko_nova_compat.nova import get_topic_queue, parse_message


class NovaResponder(Responder):
    """ Extend ``Responder`` to handle double-message nova responses and
    transform any exceptions into the format expected by the nova rpc proxy.
    """
    def __init__(self, config, msgid):
        self.config = config
        self.msgid = msgid

    def send_response(self, result, exc_info, **kwargs):
        if not self.msgid:
            return  # pragma: no cover

        # disaster avoidance serialization check
        try:
            json.dumps(result)
        except Exception:
            result = None
            exc_info = sys.exc_info()

        # failure will always json serialize
        # because we catch excs that can't be stringified
        if exc_info is not None:
            try:
                value = str(exc_info[1])
            except Exception:
                value = "[__str__ failed]"
            failure = (exc_info[0].__name__, value)
        else:
            failure = None

        conn = Connection(self.config[AMQP_URI_CONFIG_KEY])

        retry = kwargs.pop('retry', True)
        retry_policy = kwargs.pop('retry_policy', DEFAULT_RETRY_POLICY)

        with producers[conn].acquire(block=True) as producer:
            messages = [
                {'result': result, 'failure': failure, 'ending': False},
                {'result': None, 'failure': None, 'ending': True},
            ]

            for msg in messages:
                producer.publish(msg, routing_key=self.msgid, retry=retry,
                                 retry_policy=retry_policy, **kwargs)
        return result, exc_info


class NovaRpcConsumer(RpcConsumer):
    """ Extend ``RpcConsumer`` to modify the routing key, queue name, exchange
    name and handle the nova message payload.

    The result is handled by a ``NovaResponder``.
    """
    def setup(self):
        if self.queue is None:

            container = self.container
            service_name = container.service_name
            exchange_name = container.config.get('CONTROL_EXCHANGE', 'rpc')

            self.queue = get_topic_queue(exchange_name, service_name)
            self.queue_consumer.register_provider(self)

    def handle_message(self, body, message):
        try:
            routing_key = '{}.{}'.format(
                message.delivery_info['routing_key'],
                body.get('method'))

            provider = self.get_provider_for_method(routing_key)
            provider.handle_message(body, message)
        except Exception:
            msgid = body.get('_msg_id', None)
            exc_info = sys.exc_info()
            self.handle_result(message, msgid, None, exc_info)

    def handle_result(self, message, msgid, result, exc_info):
        responder = NovaResponder(self.container.config, msgid)
        result, exc_info = responder.send_response(result, exc_info)

        self.queue_consumer.ack_message(message)
        return result, exc_info


class NovaRpc(Rpc):
    """ Extend ``Rpc`` to handle the nova message payload.

    Works in combination with the ``NovaRpcConsumer``.
    """

    rpc_consumer = NovaRpcConsumer()

    def handle_message(self, body, message):

        msgid, request_ctx, _, kwargs = parse_message(body)
        args = []

        self.check_signature(args, kwargs)

        handle_result = partial(self.handle_result, message, msgid)

        context_data = request_ctx.copy()
        try:
            self.container.spawn_worker(self, args, kwargs,
                                        context_data=context_data,
                                        handle_result=handle_result)
        except ContainerBeingKilled:
            self.rpc_consumer.requeue_message(message)

    def handle_result(self, message, msgid, worker_ctx, result, exc_info):

        return self.rpc_consumer.handle_result(
            message, msgid, result, exc_info)


rpc = NovaRpc.decorator



from mock import Mock

from nameko_nova_compat import responses


def test_iter_rpcresponses():
    response_list = [
        Mock(payload={'id': 1, 'failure': False, 'ending': False}),
        Mock(payload={'id': 2, 'failure': False, 'ending': False}),
        Mock(payload={'id': 3, 'failure': False, 'ending': True}),
    ]

    iter_ = responses.iter_rpcresponses(response_list)
    ret = responses.last(iter_)

    # should be the message preceeding the `ending`
    assert ret.payload['id'] == 2


def test_iter_rpcresponses_ending_only():
    response_list = [
        Mock(payload={'id': 3, 'failure': False, 'ending': True}),
    ]

    iter_ = responses.iter_rpcresponses(response_list)

    # should not include the ending message
    assert list(iter_) == []
